# cython: language_level=3
from . import adapters, ResolveError
from .. import components


missing = object()


async def nject(*targets, **dependencies):
    """Resolve all dependencies and return the created component."""
    from buvar import context

    # create components
    cmps = components.Components()

    # add default unnamed dependencies
    for name, dep in dependencies.items():
        cmps.add(dep)

    # add current context
    cmps.push(context.current_context().index)

    # add default named dependencies
    cmps.push()
    for name, dep in dependencies.items():
        cmps.add(dep, name=name)

    # find the proper components to instantiate that class
    injected = [
        await resolve_adapter(cmps, target, name=name)
        for name, target in ((None, target) for target in targets)
    ]
    return injected


async def resolve_adapter(cmps, target, *, name=None, default=missing):
    # find in components
    try:
        component = _get_name_or_default(cmps, target, name)
        return component
    except components.ComponentLookupError:
        pass

    # try to adapt
    possible_adapters = adapters.get(target)
    if possible_adapters is None:
        if default is missing:
            raise ResolveError('Adapter not found', target)
        return default

    for adapter in possible_adapters:
        try:
            adapter_args = {
                name: await resolve_adapter(
                    cmps,
                    dependency_target,
                    name=name,
                    default=adapter.defaults.get(name, missing)
                )
                for name, dependency_target in adapter.annotations.items()
            }
        except ResolveError:
            # try next adapter
            pass
        else:
            component = await adapter.create(**adapter_args)
            # we do not use the name
            cmps.add(component)
            return component
    if default is missing:
        raise ResolveError('Adapter dependencies not found', target)
    return default


def _get_name_or_default(cmps, target, name=None):
    # find in components
    if name is not None:
        try:
            component = cmps.get(target, name=name)
            return component
        except components.ComponentLookupError:
            pass

    component = cmps.get(target, name=None)
    return component



from buvar import context


async def foo():
    ctx = context.find(str)
    context.add(ctx, 'foo_plugin')


async def plugin(include):
    await include('.bar_plugin:plugin_bar')
    context.add('foo', name='foo')

    return foo()



import pytest


@pytest.fixture
def event_loop(event_loop):
    from buvar import context, components

    cmps = components.Components()
    context.set_task_factory(cmps, loop=event_loop)
    return event_loop


@pytest.fixture(params=['resolve', 'c_resolve'])
def nject(request):
    if request.param == 'resolve':
        from buvar.di import resolve
        yield resolve.nject
    else:
        try:
            from buvar.di import c_resolve as resolve
            yield resolve.nject         # noqa: I1101
        except ImportError:
            pytest.skip(f'C extension {request.param} not available.')
            return


@pytest.mark.asyncio
async def test_di(nject):
    from buvar import context, di

    class Foo(dict):
        def __init__(self, name=None):
            super().__init__(name=name, foo=True)

    @di.register
    class Bar(dict):
        def __init__(self, bar: Foo):
            super().__init__(foo=bar, bar=True)

    class Baz(dict):
        def __init__(self, **kwargs):
            super().__init__(baz=True, **kwargs)

    class Bim:
        pass

    class Bum(dict):
        def __init__(self):
            super().__init__(bum=True)

    @di.register
    async def baz_adapter(
        bar: Bar,
        bam: Foo = 1,
        bim: Bim = 'default',
        *,
        bum: Bum,
        foo: Foo = None
    ) -> Baz:
        return Baz(foo=foo, bam=bam, bim=bim, bar=bar, bum=bum)

    context.add(Foo())
    context.add(Foo(name='bar'), name='bar')

    baz, bum = await nject(Baz, Bum, bum=Bum())
    assert isinstance(baz, Baz)
    assert baz == {
        'baz': True,
        'bam': {'foo': True, 'name': None},
        'foo': {'foo': True, 'name': None},
        'bim': 'default',
        'bum': {'bum': True},
        'bar': {'bar': True, 'foo': {'foo': True, 'name':  'bar'}},
    }
    assert bum == {'bum': True}



"""Dependency injection."""
import collections
import inspect
import itertools
import typing

import attr


class ResolveError(Exception):
    pass


@attr.s(auto_attribs=True)
class Adapter:
    callable: typing.Callable
    annotations: typing.Dict[str, typing.Any]
    defaults: typing.Dict[str, typing.Any]

    async def create(self, *args, **kwargs):
        """Create the target instance."""
        if inspect.iscoroutinefunction(self.callable):
            adapted = await self.callable(*args, **kwargs)
        else:
            adapted = self.callable(*args, **kwargs)
        return adapted

    @classmethod
    def register(cls, adapter: typing.Callable):
        """Register an adapter for later lookup."""
        # Inspect param types of class or fun.
        if inspect.isroutine(adapter):
            # __func__ indicates a classmethod
            if isinstance(adapter, classmethod):
                spec = inspect.getfullargspec(adapter.__func__)
                a_source = adapter.__func__.__annotations__
                # remove cls
                spec.args.pop(0)
            else:
                spec = inspect.getfullargspec(adapter)
                a_source = adapter.__annotations__

            # routine needs to have a return annotaion
            if "return" not in a_source:
                raise TypeError("Return type annoation missing", adapter)
            target = a_source["return"]
        elif inspect.isclass(adapter):
            target = adapter
            spec = inspect.getfullargspec(adapter)
            # remove self
            spec.args.pop(0)
        else:
            raise TypeError("Expecting a routine or a class", adapter)

        # all spec must be annotated
        annotations = {
            arg: spec.annotations[arg]
            for arg in itertools.chain(spec.args, spec.kwonlyargs)
        }
        defaults = dict(
            itertools.chain(
                zip(reversed(spec.args or []), reversed(spec.defaults or [])),
                (spec.kwonlydefaults or {}).items(),
            )
        )

        adapter_spec = cls(callable=adapter, annotations=annotations, defaults=defaults)
        adapters[target].append(adapter_spec)

        return adapter


adapters: typing.Dict[type, typing.List[Adapter]] = collections.defaultdict(list)


register = Adapter.register


try:
    # gains over 100% speed up
    from .c_resolve import nject
except ImportError:
    from .resolve import nject  # noqa: W0611



"""Provide a component registry as task context."""
import asyncio
import functools
import sys

from . import components

__all__ = ("get", "add", "find")


PY37 = sys.version_info >= (3, 7)

if PY37:

    def asyncio_current_task(loop=None):
        """Return the current task or None."""
        try:
            return asyncio.current_task(loop)
        except RuntimeError:
            # simulate old behaviour
            return None


else:
    asyncio_current_task = asyncio.Task.current_task


def default_components_context(parent_context=None):
    if parent_context is None:
        return components.Components()
    return parent_context


def task_factory(loop, coro, default=None, parent_factory=None):
    context = current_context(loop=loop)

    factory = parent_factory if parent_factory is not None else asyncio.tasks.Task
    task = factory(loop=loop, coro=coro)

    if default is None:
        default = default_components_context

    setattr(task, "context", default(context))
    return task


def current_context(loop=None):
    current_task = asyncio_current_task(loop=loop)
    context = getattr(current_task, "context", None)
    return context


def add(*args, **kwargs):
    context = current_context()
    context.add(*args, **kwargs)


def get(*args, **kwargs):
    context = current_context()
    return context.get(*args, **kwargs)


def find(*args, **kwargs):
    context = current_context()
    return context.find(*args, **kwargs)


def push():
    context = current_context()
    return context.push()


def pop():
    context = current_context()
    return context.pop()


def set_task_factory(components, *, loop=None):  # noqa: W0621
    if loop is None:
        loop = asyncio.get_event_loop()

    loop.set_task_factory(
        functools.partial(
            task_factory,
            default=lambda _: components,
            parent_factory=loop.get_task_factory(),
        )
    )



import asyncio

from buvar import context


async def plugin(load):
    import structlog
    logger = structlog.get_logger()

    fut = asyncio.Future()
    context.add(fut, name='server')
    evt_server_started = asyncio.Event()
    context.add(evt_server_started, name='server_started')

    async def server(fut):
        logger.info('Server started')
        evt_server_started.set()
        await fut
        logger.info('Server stopped')

    return server(fut)



# -*- coding: utf-8 -*-
from distutils.core import setup

package_dir = \
{'': 'src'}

packages = \
['buvar', 'buvar.di']

package_data = \
{'': ['*'], 'buvar': ['plugins/*']}

install_requires = \
['attrs>=19.1,<20.0',
 'multidict>=4.5,<5.0',
 'orjson>=2.0,<3.0',
 'structlog>=19.1,<20.0',
 'toml>=0.10,<0.11',
 'tomlkit>=0.5.3,<0.6.0']

entry_points = \
{'console_scripts': ['buvar = buvar.cli:main']}

setup_kwargs = {
    'name': 'buvar',
    'version': '0.6.1',
    'description': 'General purpose config loader',
    'long_description': None,
    'author': 'Oliver Berger',
    'author_email': 'diefans@gmail.com',
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)
